#include "planet.h"
#include <stdlib.h>

#include "constants.h"


#define NB_PLANETS  4
#define NB_FICT     4

#define DISPLAY_DISTANCE  4.0e11

const double masses[] = {M_MERCURE, M_VENUS, M_TERRE, M_MARS};
const double perihelions[] = {PER_MERCURE, PER_VENUS, PER_TERRE, PER_MARS};
const double eccentricities[] = {E_MERCURE, E_VENUS, E_TERRE, E_MARS};
const double axis_major[] = {A_MERCURE, A_VENUS, A_TERRE, A_MARS};
const uint32_t display_sun_radius = 45;
const uint32_t display_radii[] = {5, 10, 10, 20};

// fictional
const double fict_per[] = {7.80e10, 1.28e11, 1.85e11, 3.00e11};
const double fict_mas[] = {5.9e23, 7.0e23, 4.9e24, 7.0e23};
const double fict_ecc[] = {0.025, 0.005, 0.01, 0.0008};
const double fict_axm[] = {8.0e10, 1.2e11, 2.0e11, 2.9e11};
const uint32_t fict_display_radii[] = {7, 8, 14, 24};


planet_t create_planet(double mass, vec2 pos) {
	planet_t p;
	p.mass = mass;
	p.pos = pos;
	p.prec_pos = pos;  // must be changed
	return p;
}


vec2 calc_prev_position(const double star_mass, const vec2 pos, const double ecc, const double axm, const double delta_t) {
	double vel = G * star_mass * (1.0 + ecc);
	vel /= axm * (1.0 - ecc);
	vel = sqrt(vel);
	const vec2 ru = vec2_normalize(pos);
	const vec2 rp = vec2_create(-ru.y, ru.x);
	const vec2 v = vec2_mul(vel, rp);
	const vec2 prec_pos = vec2_sub(pos, vec2_mul(delta_t, v));
	return prec_pos;
}


system_t create_system(double delta_t) {
	system_t system;
	system.star = create_planet(M_SOLEIL, vec2_create_zero());
	system.nb_planets = NB_PLANETS + NB_FICT;
	system.planets = malloc(system.nb_planets * sizeof(planet_t));
	// real planets
	for (uint32_t i = 0; i < NB_PLANETS; ++i) {
		system.planets[i].mass = masses[i];
		system.planets[i].pos = vec2_create(perihelions[i], 0.0);
		system.planets[i].prec_pos = calc_prev_position(
				system.star.mass,
				system.planets[i].pos,
				eccentricities[i],
				axis_major[i],
				delta_t);
	}
	// fictional planets
	for (uint32_t i = 0; i < NB_FICT; ++i) {
		const size_t idx = i + NB_PLANETS;
		system.planets[idx].mass = fict_mas[i];
		system.planets[idx].pos = vec2_create(fict_per[i], 0.0);
		system.planets[idx].prec_pos = calc_prev_position(
				system.star.mass,
				system.planets[idx].pos,
				fict_ecc[i],
				fict_axm[i],
				delta_t);
	}
	return system;
}


void show_system(struct gfx_context_t* ctxt, system_t* system) {
	// draw sun
	const coordinates xy_sun = vec2_to_coordinates_centered(
			vec2_normalize(system->star.pos),
			ctxt->width,
			ctxt->height);
	draw_full_circle(ctxt, xy_sun.column, xy_sun.row, display_sun_radius, COLOR_YELLOW);
	// draw planets
	for (uint32_t i = 0; i < system->nb_planets; ++i) {
		const double display_distance = vec2_norm(system->planets[i].pos) / DISPLAY_DISTANCE;
		const coordinates xy = vec2_to_coordinates_centered(
				vec2_mul(display_distance, vec2_normalize(system->planets[i].pos)),
				ctxt->width,
				ctxt->height);
		if (i < NB_PLANETS) {
			draw_full_circle(ctxt, xy.column, xy.row, display_radii[i], COLOR_BLUE);
		} else {
			draw_full_circle(ctxt, xy.column, xy.row, fict_display_radii[i - NB_PLANETS], COLOR_GREEN);
		}
	}
}


void update_system(system_t* system, double delta_t) {
	for (uint32_t i = 0; i < system->nb_planets; ++i) {
		// apply the effects of the sun
		const double r_star_2 = vec2_norm_sqr(vec2_sub(system->planets[i].pos, system->star.pos));
		vec2 a = vec2_mul((G * system->star.mass) / r_star_2, vec2_normalize(system->planets[i].pos));
		// apply the effects of all other planets
		for (uint32_t k = 0; k < system->nb_planets; ++k) {
			if (k == i) continue;
			const vec2 r_ik = vec2_sub(system->planets[k].pos, system->planets[i].pos);
			const vec2 u = vec2_normalize(r_ik);  // unit vector from planet i to planet k
			// force already divided by system->planets[i].mass to get acceleration
			const double a_norm = (G * system->planets[k].mass) / vec2_norm_sqr(r_ik);
			const vec2 a_k = vec2_mul(a_norm, u);
			a = vec2_add(a, a_k);
		}
		a = vec2_mul(-1.0, a);  // correct sign: towards center
		// update position
		vec2 next_pos = vec2_mul(2.0, system->planets[i].pos);
		next_pos = vec2_sub(next_pos, system->planets[i].prec_pos);
		next_pos = vec2_add(next_pos, vec2_mul(delta_t * delta_t, a));
		system->planets[i].prec_pos = system->planets[i].pos;
		system->planets[i].pos = next_pos;
	}
}


void free_system(system_t* system) {
	free(system->planets);
}
